@@ -33,7 +33,7 @@
  * information does not come from the above mentioned specification.
  *
  * Additional references:
- *	- Intel AP-485 revision 23
+ *  - Intel AP-485 revision 25
  *    "Intel Processor Identification and the CPUID Instruction"
  *    http://developer.intel.com/design/xeon/applnots/241618.htm
  *  - DMTF Master MIF version 030621
@@ -42,6 +42,9 @@
  *  - IPMI 1.5 revision 1.1
  *    "Intelligent Platform Management Interface Specification"
  *    http://developer.intel.com/design/servers/ipmi/spec.htm
+ *  - AMD publication #20734 revision 3.04
+ *    "AMD Processor Recognition Application Note"
+ *    http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/20734.pdf
  */
 
 #include <sys/types.h>
@@ -55,6 +58,10 @@
 #include <string.h>
 #include <stdlib.h>
 
+#ifdef __ia64__
+#define USE_EFI
+#endif /* __ia64__ */
+
 #include "version.h"
 #include "types.h"
 #include "util.h"
@@ -922,8 +929,8 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 	 * the CPUID instruction. Their means are explained in table 6, but
 	 * DMI doesn't support this yet.
 	 */
-	u32 eax;
-	int cpuid=0;
+	u32 eax, edx;
+	int sig=0;
 
 	/*
 	 * This might help learn about new processors supporting the
@@ -938,7 +945,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		/*
 		 * 80386 have a different signature.
 		 */
-		printf("%sSignature: Type %X, Family %X, Major Stepping %X, Minor Stepping %X\n",
+		printf("%sSignature: Type %u, Family %u, Major Stepping %u, Minor Stepping %u\n",
 			prefix, dx>>12, (dx>>8)&0xF, (dx>>4)&0xF, dx&0xF);
 		return;
 	}
@@ -950,25 +957,39 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		 * wether the one we have here does or not. Note that this trick
 		 * works only because we know that 80486 must be little-endian.
 		 */
-		if((dx&0x0F00)==0x0400 && ((dx&0x00F0)==0x0040 || (dx&0x00F0)>=0x0070))
-			cpuid=1;
+		if((dx&0x0F00)==0x0400
+		&& ((dx&0x00F0)==0x0040 || (dx&0x00F0)>=0x0070)
+		&& ((dx&0x000F)>=0x0003))
+			sig=1;
+		else
+		{
+			printf("%sSignature: Type %u, Family %u, Model %u, Stepping %u\n",
+				prefix, (dx>>12)&0x3, (dx>>8)&0xF, (dx>>4)&0xF, dx&0xF);
+			return;
+		}
 	}
 	else if((type>=0x0B && type<=0x13) /* Intel, Cyrix */
-	|| (type>=0x18 && type<=0x1D) || type==0x1F /* AMD */
 	|| (type>=0xB0 && type<=0xB3) /* Intel */
-	|| (type>=0xB5 && type<=0xB7) /* Intel, AMD */
-	|| (type==0xB9) /* Intel */
+	|| type==0xB5 /* Intel */
+	|| type==0xB9) /* Intel */
+		sig=1;
+	else if((type>=0x18 && type<=0x1D) /* AMD */
+	|| type==0x1F /* AMD */
+	|| (type>=0xB6 && type<=0xB7) /* AMD */
 	|| (type==0x83 || type==0x84)) /* AMD 64-bit */
-		cpuid=1;
-	else if(type==0x01)
+		sig=2;
+	else if(type==0x01 || type==0x02)
 	{
 		/*
-		 * Some X86-class CPU have family "Other". In this case, we use
-		 * the version string to determine if they are known to support the
-		 * CPUID instruction.
+		 * Some X86-class CPU have family "Other" or "Unknown". In this case,
+		 * we use the version string to determine if they are known to
+		 * support the CPUID instruction.
 		 */
-		if(strcmp(version, "AMD Athlon(TM) Processor")==0)
-			cpuid=1;
+		if(strncmp(version, "Pentium III MMX", 15)==0)
+			sig=1;
+		else if(strncmp(version, "AMD Athlon(TM)", 14)==0
+		|| strncmp(version, "AMD Opteron(tm)", 15)==0)
+			sig=2;
 		else
 			return;
 	}
@@ -976,25 +997,37 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		return;
 	
 	eax=DWORD(p);
-	printf("%sSignature: Type %X, Family %X, Model %X, Stepping %X\n",
-		prefix, (eax>>12)&0x3, ((eax>>16)&0xFF0)+((eax>>8)&0x00F),
-		((eax>>12)&0xF0)+((eax>>4)&0x0F), eax&0xF);
-	if(cpuid)
+	edx=DWORD(p+4);
+	switch(sig)
 	{
-		u32 edx=DWORD(p+4);
-		
-		printf("%sFlags:", prefix);
-		if((edx&0x3FF7FDFF)==0)
-			printf(" None\n");
-		else
-		{
-			int i;
-			
-			printf("\n");
-			for(i=0; i<=31; i++)
-				if(flags[i]!=NULL && edx&(1<<i))
-					printf("%s\t%s\n", prefix, flags[i]);
-		}
+		case 1: /* Intel */
+			printf("%sSignature: Type %u, Family %u, Model %u, Stepping %u\n",
+				prefix, (eax>>12)&0x3, ((eax>>16)&0xFF0)+((eax>>8)&0x00F),
+				((eax>>12)&0xF0)+((eax>>4)&0x0F), eax&0xF);
+			break;
+		case 2: /* AMD */
+			printf("%sSignature: %s %X, %s %X, Stepping %X\n",
+				prefix,
+				((eax>>8)&0xF)==0xF?"Extended Family":"Family",
+				((eax>>8)&0xF)==0xF?(eax>>20)&0xFF:(eax>>8)&0xF,
+				((eax>>4)&0xF)==0xF?"Extended Model":"Model",
+				((eax>>4)&0xF)==0xF?(eax>>16)&0xF:(eax>>4)&0xF,
+				eax&0xF);
+			break;
+	}
+
+	edx=DWORD(p+4);
+	printf("%sFlags:", prefix);
+	if((edx&0x3FF7FDFF)==0)
+		printf(" None\n");
+	else
+	{
+		int i;
+
+		printf("\n");
+		for(i=0; i<=31; i++)
+			if(flags[i]!=NULL && edx&(1<<i))
+				printf("%s\t%s\n", prefix, flags[i]);
 	}
 }
 
@@ -1689,8 +1722,12 @@ static void dmi_on_board_devices(struct dmi_header *h, const char *prefix)
 
 	for(i=0; i<count; i++)
 	{
-		printf("%sOn Board Device Information\n",
-			prefix);
+		if(count==1)
+			printf("%sOn Board Device Information\n",
+				prefix);
+		else
+			printf("%sOn Board Device %d Information\n",
+				prefix, i+1);
 		printf("%s\tType: %s\n",
 			prefix, dmi_on_board_devices_type(p[2*i]&0x7F));
 		printf("%s\tStatus: %s\n",
@@ -3800,9 +3837,7 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 	if(myread(fd, buf, len, devmem)==-1)
 	{
 		free(buf);
-		printf("Table is unreachable, sorry. Try compiling dmidecode with -DUSE_MMAP.\n"
-			"This problem is known on the IBM T23, T30 and X30 laptops, the Fujitsu-Siemens\n"
-			"S-4582 laptop as well as IA-64 systems. If your system differ, please report\n");
+		printf("Table is unreachable, sorry. Try compiling dmidecode with -DUSE_MMAP.\n");
 		exit(1);
 	}
 #endif /* USE_MMAP */
@@ -3862,18 +3897,17 @@ int main(int argc, const char *argv[])
 	int fd, found=0;
 	off_t fp=0xF0000;
 	const char *devmem="/dev/mem";
-#ifdef __IA64__
+#ifdef USE_EFI
 	FILE *efi_systab;
 	char linebuf[64];
 #ifdef USE_MMAP
 	u32 mmoffset;
 	void *mmp;
-#else /* USE_MMAP */
-	u8 buf[0x20];
 #endif /* USE_MMAP */
-#else /* __IA64__ */
+#endif /* USE_EFI */
+#if !(defined(USE_EFI) && defined(USE_MMAP))
 	u8 buf[0x20];
-#endif /* __IA64__ */
+#endif
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4 || '\0'!=0)
 	{
@@ -3891,7 +3925,7 @@ int main(int argc, const char *argv[])
 	
 	printf("# dmidecode %s\n", VERSION);
 	
-#ifdef __IA64__
+#ifdef USE_EFI
 	if((efi_systab=fopen("/proc/efi/systab", "r"))==NULL)
 	{
 		perror("/proc/efi/systab");
@@ -3935,7 +3969,7 @@ int main(int argc, const char *argv[])
 	smbios_decode(buf, fd, argv[0], devmem);
 #endif /* USE_MMAP */
 	found++;
-#else /* __IA64__ */
+#else /* USE_EFI */
 	if(lseek(fd, fp, SEEK_SET)==-1)
 	{
 		perror(devmem);
@@ -3977,7 +4011,7 @@ int main(int argc, const char *argv[])
 			found++;
 		}
 	}
-#endif /* __IA64__ */
+#endif /* USE_EFI */
 	
 	if(close(fd)==-1)
 	{
